<?php

namespace Drupal\tmgmt_contentapi\Services;

use Drupal\Component\Utility\Xss;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Render\Markup;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\tmgmt\JobInterface;
use Drupal\tmgmt\JobQueue;
use Drupal\tmgmt\Entity\Job;
use Drupal\tmgmt_contentapi\Services\CapiDataProcessor;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\SourceFileApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateJob;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateRequestFile;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\ProviderId;
use Drupal\tmgmt_contentapi\Services\JobHelper;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\TranslationMemoryApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateRequestUpdateTM;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\tmgmt_contentapi\Services\QueueOperations;

/**
 * Service to create a job.
 */
class CreateConnectorJob {
  use StringTranslationTrait;

  /**
   * The array for content api bundle.
   *
   * @var array
   */
  protected $contentApiBundle;

  /**
   * The contains job request object.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateJob
   */
  protected $jobRequest;

  /**
   * The contains job api object.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi
   */
  protected $jobApi;

  /**
   * The contains cp Jobs.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Model\Job
   */
  protected $createdCpJob;

  /**
   * The Source file api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\SourceFileApi
   */
  protected $fileApi;

  /**
   * The request api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi
   */
  protected $transRequestApi;

  /**
   * The transfer file service.
   *
   * @var \Drupal\tmgmt_contentapi\Services\ExportJobFiles
   */
  protected $transferFiles;

  /**
   * Token to submit the job.
   *
   * @var string
   */
  protected $capiToken;

  /**
   * Token API object.
   *
   * @var array
   */
  protected $capi;

  /**
   * The stream wrapper service.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManager
   */
  protected $streamWrapperManager;

  /**
   * The filesystem service.
   *
   * @var \Drupal\Core\Datetime\DrupalDateTime
   */
  protected $drupalDateTime;

  /**
   * Message to pass when job created.
   *
   * @var string
   */
  protected $messageToPass;

  /**
   * The capi data processor service.
   *
   * @var \Drupal\tmgmt_contentapi\Services\CapiDataProcessor
   */
  protected $capiDataProcessor;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Files array to delete if exception occures.
   *
   * @var array
   */
  protected $fileArrayToDelte;

  /**
   * Zip path of current job.
   *
   * @var string
   */
  protected $zipPath;

  /**
   * Zip Archive path.
   *
   * @var obj
   */
  protected $zipArchivePath;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * The queue object used for processing jobs.
   *
   * @var object
   */
  protected $queue;

  /**
   * The job queue service.
   *
   * @var \Drupal\tmgmt\JobQueue
   */
  protected $jobQueue;

  /**
   * The queue object used for processing job items.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $itemQueue;

  /**
   * Logger Factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $logger;

  /**
   * Job helper service.
   *
   * @var \Drupal\tmgmt_contentapi\Services\JobHelper
   */
  protected $jobHelper;

  /**
   * The queue operations.
   *
   * @var \Drupal\tmgmt_contentapi\Services\QueueOperations
   */
  protected $queueOperations;

  /**
   * The queue object used for processing file.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $jobSubmissionQueue;

  /**
   * Queue name for export translation jobs to capi.
   */
  const QUEUE_NAME_EXPORT_JOBS = 'export_translation_jobs_to_capi';

  /**
   * Queue name for send files to CAPI.
   */
  const QUEUE_NAME_SEND_FILES = 'send_file_for_translation_to_capi';

  /**
   * Connctor version for CAPI.
   */
  const CONNECTOR_VERSION = 'V 9.3.5';

  /**
   * Consructor.
   *
   * @param ExportJobFiles $transferFiles
   *   Object for custom service export job files.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $streamWrapperManager
   *   Object for stream wrapper manager service.
   * @param \Drupal\tmgmt_contentapi\Services\CapiDataProcessor $capiDataProcessor
   *   Object for capi data processor service.
   * @param \Drupal\tmgmt\JobQueue $jobQueue
   *   Job queue object.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger
   *   Service for logger factory.
   * @param \Drupal\tmgmt_contentapi\Services\JobHelper $jobHelper
   *   Service for job helper.
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   Service for queue factory.
   * @param \Drupal\tmgmt_contentapi\Services\QueueOperations $queueOperations
   *   Service for queue operations.
   */
  public function __construct(
    ExportJobFiles $transferFiles,
    StreamWrapperManagerInterface $streamWrapperManager,
    CapiDataProcessor $capiDataProcessor,
    JobQueue $jobQueue,
    LoggerChannelFactoryInterface $logger,
    JobHelper $jobHelper,
    QueueFactory $queueFactory,
    QueueOperations $queueOperations,
  ) {

    $this->transferFiles = $transferFiles;
    $this->streamWrapperManager = $streamWrapperManager;
    $this->capiDataProcessor = $capiDataProcessor;
    $this->jobQueue = $jobQueue;
    $this->logger = $logger;
    $this->jobHelper = $jobHelper;
    $this->queueFactory = $queueFactory;
    $this->queueOperations = $queueOperations;
    // Set default values for variables.
    $this->jobSubmissionQueue = $this->queueFactory->get(self::QUEUE_NAME_EXPORT_JOBS);
    $this->itemQueue = $this->queueFactory->get(self::QUEUE_NAME_SEND_FILES);
    $this->contentApiBundle = [];
    $this->jobRequest = NULL;
    $this->jobApi = new JobApi();
    $this->messageToPass = $this->t('Exported files can be downloaded here: <br/>');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
          $container->get('tmgmt_contentapi.export_job'),
          $container->get('stream_wrapper_manager'),
          $container->get('tmgmt_contentapi.capi_data_processor'),
          $container->get('tmgmt.queue'),
          $container->get('logger.factory'),
          $container->get('tmgmt_contentapi.job_helper'),
          $container->get('queue'),
          $container->get('tmgmt_contentapi.queue_operations'),
      );
  }

  /**
   * Summary of genrateJobRequst.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   *
   * @return \Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateJob
   *   Request job object.
   */
  public function genrateJobRequst(JobInterface $job) {
    $source_lang = $job->getRemoteSourceLanguage();
    $target_lang = $job->getRemoteTargetLanguage();
    $shouldquote = (bool) ($job->getSetting('capi-settings')['quote']['is_quote']);
    $capijobsettings = $job->getSetting("capi-settings");
    $jobarray = [
      'job_name' => $this->jobHelper->getJobLabelNoSpeChars($job),
      'description' => $capijobsettings["description"] ?? NULL,
      'po_reference' => $capijobsettings["po_reference"] ?? NULL,
      'due_date' => $capijobsettings["due_date"] ?? NULL,
      'should_quote' => $shouldquote,
      'source_lang_code' => $source_lang,
      'target_lang_code' => $target_lang,
      'connector_name' => 'Lionbridge Connector for Drupal 9 and 10',
      'connector_version' => self::CONNECTOR_VERSION,
    ];
    if (isset($capijobsettings["custom_data"]) && $capijobsettings["custom_data"] !== "") {
      $job['custom_data'] = $capijobsettings["custom_data"];
    }
    $jobrequest = new CreateJob($jobarray);
    return $jobrequest;
  }

  /**
   * Function to create cp jobs.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   */
  public function createCpJobs(JobInterface $job) {
    // Create cp job.
    $this->jobRequest = $this->genrateJobRequst($job);
    $translator = $job->getTranslator();
    $this->capiToken = \DRUPAL::service('tmgmt_contentapi.capi_details')->getCapiToken($translator);
    $this->createdCpJob = $this->jobApi->jobsPost($this->capiToken, $this->jobRequest);
  }

  /**
   * Function to transfer files to cpa.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   */
  public function uploadTransferFilesToCpa(JobInterface $job) {

    // Save job to make sure all settings are available later when job queue and item queue are processed.
    $job->save();
    $upload_info = [
      'item_id' => $job->id(),
      'job' => $job,
    ];
    // Check if job is already in queue.
    // Get existing queue items.
    $queue_items = $this->queueOperations->getQueueItems(self::QUEUE_NAME_EXPORT_JOBS, 'item_id');
    // Check if job already exists in queue or if already processed.
    if (in_array($job->id(), $queue_items) || !$job->isUnprocessed()) {
      $this->logger->get("TMGMT_CONTENTAPI_JOB_POST")->notice(
        'Job already exists in queue to send it for submission job ids: @jobids ',
        [
          '@jobids' => $job->id(),
        ]
      );
      return;
    }
    $this->jobSubmissionQueue->createItem($upload_info);
    $remainingJobCount = $this->jobQueue->count();
    $workers = 1;
    // If this is last jobs in process then start queue item processing.
    // Execut queue using only one worker. Multiple workers leads to item stuck in queue.
    // If job item alread exist in queue it means queue worker is already started.
    // Then no need to create new queue worker.
    if ($remainingJobCount <= $workers) {
      $this->queueOperations->processQueue(self::QUEUE_NAME_EXPORT_JOBS);
    }
  }

  /**
   * Function to batch submit files to CAPI.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   * @param bool $is_translation_memory
   *   Check if job is for translation memory.
   */
  public function addFilesToQueueForSubmission(JobInterface $job, $is_translation_memory = FALSE) {
    // Check if get translator is not null.
    try {
      $this->contentApiBundle = [];
      $transfer_files = [];
      $this->createdCpJob = NULL;
      if (!$is_translation_memory) {
        $this->createCpJobs($job);
        $capi_job_id = $this->createdCpJob->getJobId();
      }
      else {
        $capi_job_id = $this->jobHelper->getCpJobIdfromLocJob($job);
      }
      $transfer_files = $this->transferFiles->jobFilesToTransfer($job, $is_translation_memory);
      // Add transfer file array to state variable.
      // To avoid it adding in every item in queue.
      $this->queueOperations->addValueToStatevariable($capi_job_id . '_' . $job->id() . '_transfer_files', $transfer_files);
      // Get total items for this job.
      $total_items = count($transfer_files);
      // Add item info in queue for processing.
      foreach ($transfer_files as $id => $tmpfile) {
        $data = [];
        $data["filename"] = $tmpfile->getFilename();
        $stmrg = $this->streamWrapperManager->getViaUri($tmpfile->getFileUri());
        $extpath = $stmrg->realpath();
        $zip = new \ZipArchive();
        if ($zip->open($extpath) === TRUE) {
          $zip->extractTo($this->transferFiles->allFilesPath);
          $zip->close();
        }
        $data["filetype"] = $tmpfile->getMimeType();
        $stmrg = $this->streamWrapperManager->getViaUri($tmpfile->getFileUri());
        $extpath = $stmrg->realpath();
        $jobpath = $job->getTranslator()->getSetting('transfer-settings') ? $extpath : $this->transferFiles->allFilesPath . "/" . $data["filename"];

        $item_info = [
          'id' => $id,
          'jobid' => $job->id(),
          'capi_job_id' => $capi_job_id,
          'translator' => $job->getTranslator(),
          'is_translation_memory' => $is_translation_memory,
          'getRemoteSourceLanguage' => $job->getRemoteSourceLanguage(),
          'getRemoteTargetLanguage' => $job->getRemoteTargetLanguage(),
          'data' => $data,
          'extpath' => $extpath,
          'jobpath' => $jobpath,
          'allFilesPath' => $this->transferFiles->allFilesPath,
          'zipPath' => $this->transferFiles->zipPath,
          'total_items' => $total_items,
        ];
        $this->itemQueue->createItem($item_info);
      }
      $this->logger->get("TMGMT_CONTENTAPI_JOB_POST")->notice(
        'File added in queue to send it for submission job ids: @jobids ',
        [
          '@jobids' => $job->id() . '_' . $capi_job_id,
        ]
      );
      // Start send files worker only when all job files are added in queue.
      if ($this->jobSubmissionQueue->numberOfItems() > 1) {
        return;
      }
      $this->queueOperations->processQueue(self::QUEUE_NAME_SEND_FILES);
    }
    catch (\Exception $ex) {
      if (!$is_translation_memory) {
        if (isset($this->createdCpJob)) {
          $capi_job_id = $this->createdCpJob->getJobId();
        }
      }
      else {
        $capi_job_id = $this->jobHelper->getCpJobIdfromLocJob($job);
      }
      // Release job from queue if any exception occurs.
      $this->cleanUpJobOnFailure($job->id(), $capi_job_id, $this->transferFiles->allFilesPath, $this->transferFiles->zipPath, $ex);
      // Log the error.
      $this->logger->get(channel: 'TMGMT_CONTENTAPI_JOB_POST')->error(
            'Issue while processing item : @joblabel, Job: @job, CAPI jobid: @capijobid Error: @errormessage.',
            context: [
              '@joblabel' => $job->label(),
              '@job' => $job->id(),
              '@capijobid' => $capi_job_id,
              '@errormessage' => $ex->getMessage(),
            ]
            );
    }
    // Free memory to avoid memory leak.
    unset($job);
    unset($this->contentApiBundle);
    unset($transfer_files);
    unset($this->createdCpJob);
    // Collect garbage.
    gc_collect_cycles();
  }

  /**
   * Function to post files to CAPI.
   *
   * @param array $item_info
   *   Array containing item information.
   */
  public function postFilesToCapi($item_info) {
    // Get all the information from item info array.
    $id = $item_info['id'];
    $jobid = $item_info['jobid'];
    $capi_job_id = $item_info['capi_job_id'];
    $translator = $item_info['translator'];
    $is_translation_memory = $item_info['is_translation_memory'];
    $getRemoteSourceLanguage = $item_info['getRemoteSourceLanguage'];
    $getRemoteTargetLanguage = $item_info['getRemoteTargetLanguage'];
    $data = $item_info['data'];
    $extpath = $item_info['extpath'];
    $jobpath = $item_info['jobpath'];
    $allFilesPath = $item_info['allFilesPath'];
    $zipPath = $item_info['zipPath'];
    $total_items = $item_info['total_items'];

    // Get capi token.
    $this->capiToken = \DRUPAL::service('tmgmt_contentapi.capi_details')->getCapiToken($translator);
    // Create object for file api and request api.
    $fileApi = new SourceFileApi();
    $transRequestApi = new RequestApi();
    // Process the file.
    try {
      $filerequest = new \SplFileObject($extpath);
      // File upload to TUS CLINET.
      $contentapitmpfile = $fileApi->jobsJobIdUploadPost($this->capiToken, $capi_job_id, $data["filename"], $data["filetype"], $filerequest);
      $fileApi->uploadFile($contentapitmpfile, $jobpath);
      // If request is for translation memory then update translation memory.
      // Else continue with file upload for job creation.
      if ($is_translation_memory) {
        $tmupdateapi = new TranslationMemoryApi();
        $tmupdaterequest = new CreateRequestUpdateTM(
          [
            'file_id' => $contentapitmpfile->getfms_file_id(),
            'source_native_language_code' => $getRemoteSourceLanguage,
            'target_native_language_code' => $getRemoteTargetLanguage,
          ]
          );
        $tmupdateapi->jobsJobIdTmUpdatefilePut($this->capiToken, $capi_job_id, $tmupdaterequest);
      }
      else {
        $request = new CreateRequestFile([
          "request_name" => $jobid . "_" . $id . "_" . $getRemoteSourceLanguage . "-" . $getRemoteTargetLanguage,
          "source_native_id" => $jobid . "_" . $id,
          "source_native_language_code" => $getRemoteSourceLanguage,
          "target_native_language_codes" => [$getRemoteTargetLanguage],
          "file_id" => $contentapitmpfile['fms_file_id'],
        ]);
        $contentapitmprequst = $transRequestApi->jobsJobIdRequestsAddfilePost($this->capiToken, $capi_job_id, $request);
        // Add response table for further processing.
        $this->queueOperations->addApiResponse($capi_job_id . '_' . $jobid, $id, $contentapitmprequst);
      }
      // Log the file post success.
      $this->logger->get("TMGMT_CONTENTAPI_FILE_POST_SUCCESS")->notice(
        'jobids: @jobids - File posted successfully for file:@filename ',
        [
          '@jobids' => $jobid . '_' . $capi_job_id,
          '@filename' => $data["filename"],
        ]);
      // Check if all files are processed then submit job for processing.
      $total_processed_items = $this->queueOperations->getApiResponsesCountByJobKey($capi_job_id . '_' . $jobid);
      if ($total_processed_items == $total_items) {
        $this->logger->get("TMGMT_CONTENTAPI_JOB_POST_ITEM_COUNT")->notice(
          'jobids: @jobids - total_items :@total_items - processed_items: @processed_items ',
          [
            '@jobids' => $jobid . '_' . $capi_job_id,
            '@total_items' => $total_items,
            '@processed_items' => $total_processed_items,
          ]);
        $lock = \Drupal::lock();
        if ($lock->acquire("final_submit_lock_$capi_job_id")) {
          try {
            $this->submitJobPostUpload($jobid, $capi_job_id, $allFilesPath, $zipPath);
          } finally {
            $lock->release("final_submit_lock_$capi_job_id");
          }
        }
      }
    }
    catch (\Exception $ex) {
      $this->cleanUpJobOnFailure($item_info['jobid'], $item_info['capi_job_id'], $item_info['allFilesPath'], $item_info['zipPath'], $ex);
      $this->logger->get("TMGMT_CONTENTAPI_FILE_POST_FAILED")->notice(
        'jobids: @jobids - Failed File posting for file:@filename, Error: @errormessage ',
        [
          '@jobids' => $item_info['jobid'] . '_' . $item_info['capi_job_id'],
          '@filename' => $data["filename"],
          '@errormessage' => $ex->getMessage(),
        ]
            );
    }
    unset($contentapitmpfile);
    unset($contentapitmprequst);
    unset($request);
    unset($filerequest);
  }

  /**
   * Function to submit job to content API.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   */
  public function submitJobToContentApi(JobInterface $job) {
    try {
      $this->drupalDateTime = new DrupalDateTime('now', 'UTC');
      $this->uploadTransferFilesToCpa($job);
      $this->transferFiles->refFilesToTransfer($job);
      // Generate message which user can see once job added to queue for further processing.
      $messageToPass = $this->t('Job(s) creation is in progress and is running in the background. You can view the progress and updated status on the Job Overview page.');
      \Drupal::messenger()->addMessage(Markup::create($messageToPass));
    }
    catch (\Exception $ex) {
      $this->logger->get('TMGMT_CONTENTAPI_JOB_POST')->error(
            'Issue while creating job : @joblabel, Error: @errormessage.',
            [
              '@joblabel' => $job->label(),
              '@errormessage' => $ex->getMessage(),
            ]
            );
    }
  }

  /**
   * Function to submit translation memory to content API.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   */
  public function updateTranslationMemory(JobInterface $job) {
    // If job state is not finished then return.
    if ($job->getState() != JobInterface::STATE_FINISHED) {
      return;
    }
    // Submit translation memory to content api.
    $translator = $job->getTranslator();
    $this->capiToken = \DRUPAL::service('tmgmt_contentapi.capi_details')->getCapiToken($translator);
    // If job submission for translation memory.
    $this->addFilesToQueueForSubmission($job, TRUE);
    $this->transferFiles->deleteProcessedFiles();
    $job->submitted($this->t("Job sent to update translation memory!"));
    $messageToPass = $this->t('TM update request successfully submitted.<br/>Exported files can be downoaded here:<br/>');
    $zipDownload = '';
    if (isset($this->transferFiles->zipPath)) {
      $zipDownload = \Drupal::service('file_url_generator')->generateAbsoluteString($this->transferFiles->zipPath);
    }
    $messageTopass .= '<a href="' . $zipDownload . '">' . Xss::filter($this->jobHelper->getJobLabelNoSpeChars($job)) . '</a>';
    \Drupal::messenger()->addMessage(Markup::create($messageToPass));
  }

  /**
   * Function to submit job to content API.
   *
   * @param int $jobid
   *   Job under process.
   * @param string $capi_job_id
   *   Capi job id.
   * @param string $allFilesPath
   *   All files path.
   * @param string $zipPath
   *   Zip path.
   */
  public function submitJobPostUpload(int $jobid, string $capi_job_id, $allFilesPath, $zipPath) {
    try {
      // Load job object.
      $job = Job::load($jobid);
      $pr = $job->getSetting("capi-settings")["provider"];
      $jid = $capi_job_id;
      $prmodel = new ProviderId(['provider_id' => $pr]);
      $jobsubmitresult = $this->jobApi->jobsJobIdSubmitPut($this->capiToken, $jid, $prmodel);
      if ($jobsubmitresult) {
        // Add content api bundle to job's settings.
        $contentApiBundle = $this->queueOperations->getApiResponsesByJobKey($capi_job_id . '_' . $jobid);
        $this->logger->get("TMGMT_CONTENTAPI_JOB_POST_CONTENT_BUNDEL_COUNT")->notice('Count: @count ', ['@count' => count($contentApiBundle)]);
        $this->jobHelper->addCpaSettingsToJob($job, serialize($contentApiBundle));
        // Set message if quote is set else success.
        if ($job->getSetting('capi-settings')['quote']['is_quote']) {
          $job->submitted($this->t("This job was submitted for a quote. To submit your job for processing, you must log into your translation provider's system to approve this quote."));
        }
        else {
          $job->submitted($this->t("Job sent to provider!"));
        }
        // Store job details job status table for further processing.
        $this->capiDataProcessor->jobToinsertInProcessor($job, $jobsubmitresult->getProviderId());
        // Delete transfer files.
        $transfer_files = $this->queueOperations->getvalueFromStatevariable($capi_job_id . '_' . $jobid . '_transfer_files');
        $this->transferFiles->deleteProcessedFiles($transfer_files[0]);
        // Delete all temporary state variables and response data.
        $this->queueOperations->deleteApiResponsesByJobKey($capi_job_id . '_' . $jobid);
        $this->queueOperations->deleteStateVariable($capi_job_id . '_' . $jobid . '_transfer_files');
        // Log the job sent successfully.
        $this->logger->get("TMGMT_CONTENTAPI_JOB_POST")->notice('Job sent successfully job ids: @jobids ', ['@jobids' => $job->id() . '_' . $jid]);
      }
    }
    catch (\Throwable $ex) {
      // Log the reson for error.
      $this->logger->get('TMGMT_CONTENTAPI_JOB_POST')->error(
            'Failed to process job, please resubmit the same job: @joblabel, jobid: @jobids, language: @languages Error: @errormessage.',
            context: [
              '@joblabel' => $job->label(),
              '@jobids' => $job->id(),
              '@languages' => $job->getRemoteSourceLanguage() . ' to ' . $job->getRemoteTargetLanguage(),
              '@errormessage' => $ex->getMessage(),
            ]
            );
      $this->cleanUpJobOnFailure($jobid, $capi_job_id, $allFilesPath, $zipPath, $ex);
    }
  }

  /**
   * If exception occurs, clean up everything: delete exported files, cancel job in CA if any.
   *
   * @param int $jobid
   *   Job under process.
   * @param string $capi_job_id
   *   Capi job id.
   * @param string $allFilesPath
   *   All files path.
   * @param string $zipPath
   *   Zip path.
   * @param \Exception $ex
   *   Exception object.
   */
  public function cleanUpJobOnFailure(int $jobid, string $capi_job_id, string $allFilesPath, string $zipPath, \Throwable $ex) {
    $this->logger->get('TMGMT_CONTENTAPI_JOB_POST')->error(
      'Cleaning up process started for job: @jobids due to error: @errormessage. ',
      context: [
        '@jobids' => $jobid . '_' . $capi_job_id,
        '@errormessage' => $ex->getMessage(),
      ]
      );
    // Step 1: Delete all files from temporary directory.
    $job = Job::load($jobid);
    $transfer_files = $this->queueOperations->getvalueFromStatevariable($capi_job_id . '_' . $jobid . '_transfer_files');
    // If exception occurs, clean up everything: delete exported files, cancel job in CA if any.
    if (!empty($transfer_files[0])) {
      foreach ($transfer_files[0] as $tempfile) {
        if (file_exists($tempfile->getFileUri())) {
          \Drupal::service('file_system')->delete($tempfile->getFileUri());
        }
      }
    }

    // Step 2: Keep job in unprocessed state but delete job generated in CAPI and records in custom table.
    try {
      if (isset($capi_job_id)) {
         $this->jobApi->jobsJobIdDelete($this->capiToken, $capi_job_id);
         // Delete entries related to this job from processor table.
         $this->capiDataProcessor->deleteProcessorRecords($capi_job_id, $jobid);
         // Delete entries related to this job from message table.
         $this->capiDataProcessor->deleteMessageRecords($jobid);
         // Keep job in unprocessed state.
         $job->setState(JobInterface::STATE_UNPROCESSED);
      }
    }
    catch (\Exception $exception) {
      $msg = $this->t('Tried to delete Job in Content API, but failed: @message', ['@message' => $exception->getMessage()]);
      if (strlen($msg) > 200) {
        $msg = substr($msg, 0, 200);
      }
      \Drupal::messenger()->addMessage($msg, "error");
    }

    // Step 3: Delete zip files.
    // Get length of zip path.
    if (strlen($zipPath) < 250) {
      $zipfileobj = $this->jobHelper->createFileObject($zipPath);
      if ($zipfileobj->getFileUri() != NULL) {
        \Drupal::service('file_system')->delete($zipfileobj->getFileUri());
      }
    }
    if ($this->transferFiles->allFilesPath != NULL) {
      \Drupal::service('file_system')->deleteRecursive($allFilesPath);
    }

    // Step 4: Delete all queue items(files) for this job which queued for upload.
    $queue_items = $this->queueOperations->getQueueItems(self::QUEUE_NAME_SEND_FILES, 'jobid');
    $current_job_id = [$jobid];
    $specific_job_array = array_intersect($queue_items, $current_job_id);
    foreach ($specific_job_array as $item_id => $item) {
      $this->queueOperations->deleteItemFromQueue($item_id);
    }

    // Step 5: Finally Delete all temporary state variables and rows from custom table.
    $this->queueOperations->deleteApiResponsesByJobKey($capi_job_id . '_' . $jobid);
    $this->queueOperations->deleteStateVariable($capi_job_id . '_' . $jobid . '_transfer_files');

    // Step 6: Set exact error occured while job creation.
    $msg = $ex->getMessage();
    if (strlen($msg) > 200) {
      $msg = substr($msg, 0, 200);
    }
    \Drupal::messenger()->addMessage($job->label() . $msg, 'error');

  }

}
